@@ -1,7 +1,7 @@
 use std::fmt;
 use std::iter::{self, FromIterator};
 use std::marker::PhantomData;
-use std::ops::Deref;
+use std::ops::{Add, Deref, Mul};
 
 use afarray::{Array, ArrayExt, ArrayInstance, CoordBlocks, Coords, Offsets};
 use arrayfire as af;
@@ -17,15 +17,16 @@ use tc_btree::Node;
 use tc_error::*;
 use tc_transact::fs::{BlockId, CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{Transact, Transaction, TxnId};
-use tc_value::{Number, NumberClass, NumberInstance, NumberType};
+use tc_value::{Float, Number, NumberClass, NumberInstance, NumberType};
 use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use crate::stream::{Read, ReadValueAt};
-use crate::transform;
-use crate::{coord_bounds, Bounds, Coord, Schema, Shape, TensorAccess, TensorType};
+use crate::{
+    coord_bounds, transform, Bounds, Coord, FloatType, Schema, Shape, TensorAccess, TensorType,
+};
 
 use super::access::BlockListTranspose;
-use super::{DenseAccess, DenseAccessor, DenseWrite, MEBIBYTE, PER_BLOCK};
+use super::{array_err, div_ceil, DenseAccess, DenseAccessor, DenseWrite, MEBIBYTE, PER_BLOCK};
 
 /// The size of a dense tensor block on disk, in bytes (1 mebibyte + 5 bytes overhead).
 const BLOCK_SIZE: usize = MEBIBYTE + 5;
@@ -63,6 +64,53 @@ where
     pub async fn constant(file: FD, txn_id: TxnId, shape: Shape, value: Number) -> TCResult<Self> {
         debug!("BlockListFile::constant {} {}", shape, value);
 
+        let value_clone = value.clone();
+        let generator = |length| Array::constant(value_clone.clone(), length);
+        Self::from_blocks_generator(file, txn_id, shape, value.class(), generator).await
+    }
+
+    /// Construct a new `BlockListFile` with the given [`Shape`], with a random normal distribution.
+    pub async fn random_normal(
+        file: FD,
+        txn_id: TxnId,
+        shape: Shape,
+        dtype: FloatType,
+        mean: Float,
+        std: Float,
+    ) -> TCResult<Self> {
+        debug!("BlockListFile::random_normal {}", shape);
+        let generator = |length| {
+            // TODO: impl Add<Number> and Mul<Number> for Array
+            let std = Array::constant(std.into(), length);
+            let mean = Array::constant(mean.into(), length);
+            Array::random_normal(dtype, length).mul(&std).add(&mean)
+        };
+
+        Self::from_blocks_generator(file, txn_id, shape, dtype.into(), generator).await
+    }
+
+    /// Construct a new `BlockListFile` with the given [`Shape`], with a random normal distribution.
+    pub async fn random_uniform(
+        file: FD,
+        txn_id: TxnId,
+        shape: Shape,
+        dtype: FloatType,
+    ) -> TCResult<Self> {
+        debug!("BlockListFile::random_normal {}", shape);
+        let generator = |length| Array::random_uniform(dtype, length);
+        Self::from_blocks_generator(file, txn_id, shape, dtype.into(), generator).await
+    }
+
+    async fn from_blocks_generator<G>(
+        file: FD,
+        txn_id: TxnId,
+        shape: Shape,
+        dtype: NumberType,
+        generator: G,
+    ) -> TCResult<Self>
+    where
+        G: Fn(usize) -> Array + Send + Copy,
+    {
         if !file.is_empty(txn_id).await? {
             return Err(TCError::unsupported(
                 "cannot create new tensor: file is not empty",
@@ -71,30 +119,14 @@ where
 
         let size = shape.size();
 
-        let value_clone = value.clone();
-        let blocks = (0..(size / PER_BLOCK as u64))
-            .map(move |_| Ok(Array::constant(value_clone.clone(), PER_BLOCK)));
+        let blocks = (0..(size / PER_BLOCK as u64)).map(move |_| Ok(generator(PER_BLOCK)));
 
         let trailing_len = (size % PER_BLOCK as u64) as usize;
         if trailing_len > 0 {
-            let blocks = blocks.chain(iter::once(Ok(Array::constant(value.clone(), trailing_len))));
-            Self::from_blocks(
-                file,
-                txn_id,
-                Some(shape),
-                value.class(),
-                stream::iter(blocks),
-            )
-            .await
+            let blocks = blocks.chain(iter::once(Ok(generator(trailing_len))));
+            Self::from_blocks(file, txn_id, Some(shape), dtype, stream::iter(blocks)).await
         } else {
-            Self::from_blocks(
-                file,
-                txn_id,
-                Some(shape),
-                value.class(),
-                stream::iter(blocks),
-            )
-            .await
+            Self::from_blocks(file, txn_id, Some(shape), dtype, stream::iter(blocks)).await
         }
     }
 
@@ -106,20 +138,28 @@ where
         dtype: NumberType,
         blocks: S,
     ) -> TCResult<Self> {
+        let bytes_per_element = dtype.size();
         let size = blocks
             .enumerate()
             .map(|(i, r)| r.map(|block| (BlockId::from(i), block)))
             .map_ok(|(id, block)| {
-                let len = block.len() as u64;
-                file.create_block(txn_id, id, block, BLOCK_SIZE)
-                    .map_ok(move |_| len)
+                let len = block.len();
+                file.create_block(txn_id, id, block, len * bytes_per_element)
+                    .map_ok(move |_| len as u64)
             })
             .try_buffer_unordered(num_cpus::get())
             .try_fold(0u64, |block_len, size| future::ready(Ok(size + block_len)))
             .await?;
 
         let shape = if let Some(shape) = shape {
-            if shape.size() != size {
+            if shape.size() < size {
+                return Err(TCError::unsupported(format!(
+                    "dense tensor with shape {} requires {} elements but found {}",
+                    shape,
+                    shape.size(),
+                    size
+                )));
+            } else if shape.size() > size {
                 return Err(TCError::unsupported(format!(
                     "dense tensor of shape {} requires {} elements but found {}--this could indicate a divide-by-zero error",
                     shape,
@@ -164,8 +204,8 @@ where
             return Err(TCError::unsupported(format!(
                 "DenseTensor of shape {} requires {} values, found {}",
                 shape,
+                shape.size(),
                 size,
-                shape.size()
             )));
         }
 
@@ -208,6 +248,11 @@ where
         Box::pin(blocks)
     }
 
+    /// Borrow this `BlockListFile`'s underlying `File`.
+    pub(super) fn file(&self) -> &FD {
+        &self.file
+    }
+
     /// Sort the elements in this `BlockListFile`.
     pub async fn merge_sort(&self, txn_id: TxnId) -> TCResult<()> {
         let num_blocks = div_ceil(self.size(), PER_BLOCK as u64);
@@ -217,29 +262,36 @@ where
             let block_id = BlockId::from(0u64);
             let mut block = self.file.write_block(txn_id, block_id).await?;
 
-            block.sort(true)?;
+            block.sort(true).map_err(array_err)?;
             return Ok(());
         }
 
-        for block_id in 0..(num_blocks - 1) {
-            let next_block_id = BlockId::from(block_id + 1);
-            let block_id = BlockId::from(block_id);
+        loop {
+            let mut sorted = true;
+            for block_id in 0..(num_blocks - 1) {
+                let next_block_id = BlockId::from(block_id + 1);
+                let block_id = BlockId::from(block_id);
 
-            let left = self.file.write_block(txn_id, block_id);
+                let left = self.file.write_block(txn_id, block_id);
 
-            let right = self.file.write_block(txn_id, next_block_id);
+                let right = self.file.write_block(txn_id, next_block_id);
 
-            let (mut left, mut right) = try_join!(left, right)?;
+                let (mut left, mut right) = try_join!(left, right)?;
 
-            let mut block = Array::concatenate(&left, &right);
-            block.sort(true)?;
+                let mut block = Array::concatenate(&left, &right);
+                block.sort(true).map_err(array_err)?;
 
-            let (left_sorted, right_sorted) = block.split(PER_BLOCK)?;
-            *left = left_sorted;
-            *right = right_sorted;
+                let (left_sorted, right_sorted) = block.split(PER_BLOCK).map_err(array_err)?;
+                sorted = sorted && left.deref() == &left_sorted && right.deref() == &right_sorted;
+
+                *left = left_sorted;
+                *right = right_sorted;
+            }
+
+            if sorted {
+                return Ok(());
+            }
         }
-
-        Ok(())
     }
 
     async fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
@@ -260,7 +312,7 @@ where
 
         (*block)
             .set_value(offset as usize, value)
-            .map_err(TCError::from)
+            .map_err(array_err)
     }
 
     async fn overwrite<B: DenseAccess<FD, FS, D, T>>(&self, txn: T, value: B) -> TCResult<()> {
@@ -453,11 +505,11 @@ where
                             af::sum_all(&af::eq(&block_offsets, af_block_id.deref(), true));
                         let end = start + len as usize;
                         let indices = indices.slice(start, end);
-                        let array = array.slice(start, end).map_err(TCError::from)?;
+                        let array = array.slice(start, end).map_err(array_err)?;
 
                         let mut block = self.file.write_block(txn_id, block_id.into()).await?;
 
-                        block.set(&indices, &array)?;
+                        block.set(&indices, &array).map_err(array_err)?;
 
                         start = end;
                     }
@@ -474,10 +526,8 @@ where
         debug!("BlockListFile::write_value {} at {}", value, bounds);
         self.shape().validate_bounds(&bounds)?;
 
-        if bounds.len() == self.ndim() {
-            if let Some(coord) = bounds.as_coord() {
-                return self.write_value_at(txn_id, coord, value).await;
-            }
+        if let Some(coord) = bounds.as_coord(self.shape()) {
+            return self.write_value_at(txn_id, coord, value).await;
         }
 
         bounds.normalize(self.shape());
@@ -502,7 +552,10 @@ where
                         let mut block = file.write_block(txn_id, block_id).await?;
 
                         let value = Array::constant(value, (new_start - start) as usize);
-                        (*block).set(&block_offsets.into(), &value)?;
+                        (*block)
+                            .set(&block_offsets.into(), &value)
+                            .map_err(array_err)?;
+
                         start = new_start;
                     }
 
@@ -1075,15 +1128,6 @@ impl<FD, FS, D, T> fmt::Display for BlockListFileSlice<FD, FS, D, T> {
     }
 }
 
-#[inline]
-fn div_ceil(l: u64, r: u64) -> u64 {
-    if l % r == 0 {
-        l / r
-    } else {
-        (l / r) + 1
-    }
-}
-
 fn block_offsets(
     indices: &ArrayExt<u64>,
     offsets: &ArrayExt<u64>,
