@@ -2,33 +2,35 @@ use std::fmt;
 use std::marker::PhantomData;
 use std::ops::{Add, Div, Mul, Sub};
 
-use afarray::{Array, ArrayInstance};
+use afarray::{Array, ArrayInstance, CoordBlocks};
 use arrayfire as af;
 use async_trait::async_trait;
 use destream::{de, en};
 use futures::future::{self, TryFutureExt};
 use futures::stream::{Stream, StreamExt, TryStreamExt};
 use log::{debug, warn};
-use safecast::{AsType, CastFrom};
+use safecast::{AsType, CastFrom, CastInto};
 
 use tc_btree::Node;
 use tc_error::*;
-use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
+use tc_transact::fs::{CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
-use tc_value::{FloatType, Number, NumberClass, NumberInstance, NumberType};
+use tc_value::{
+    Float, FloatType, Number, NumberClass, NumberInstance, NumberType, Trigonometry, UIntType,
+};
 use tcgeneric::{Instance, TCBoxTryFuture, TCBoxTryStream};
 
 use super::sparse::{DenseToSparse, SparseTensor};
 use super::stream::{Read, ReadValueAt};
 use super::{
-    Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorBooleanConst,
-    TensorCompare, TensorCompareConst, TensorDualIO, TensorIO, TensorInstance, TensorMath,
-    TensorMathConst, TensorPersist, TensorReduce, TensorTransform, TensorType, TensorUnary,
-    ERR_COMPLEX_EXPONENT,
+    tile, trig_dtype, Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean,
+    TensorBooleanConst, TensorCompare, TensorCompareConst, TensorDiagonal, TensorDualIO, TensorIO,
+    TensorIndex, TensorInstance, TensorMath, TensorMathConst, TensorPersist, TensorReduce,
+    TensorTransform, TensorTrig, TensorType, TensorUnary, ERR_COMPLEX_EXPONENT,
 };
 
 use access::*;
-pub use access::{BlockListSparse, DenseAccess, DenseAccessor, DenseWrite};
+pub use access::{arg_sort, BlockListSparse, DenseAccess, DenseAccessor, DenseWrite};
 pub use file::BlockListFile;
 
 mod access;
@@ -109,9 +111,12 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
 {
     /// Create a new `DenseTensor` with the given [`Schema`].
     pub async fn create(file: FD, schema: Schema, txn_id: TxnId) -> TCResult<Self> {
+        schema.validate("create Dense")?;
+
         let Schema { shape, dtype } = schema;
         BlockListFile::constant(file, txn_id, shape, dtype.zero())
             .map_ok(Self::from)
@@ -123,7 +128,13 @@ where
     where
         Shape: From<S>,
     {
-        BlockListFile::constant(file, txn_id, shape.into(), value)
+        let schema = Schema {
+            shape: shape.into(),
+            dtype: value.class(),
+        };
+        schema.validate("create Dense constant")?;
+
+        BlockListFile::constant(file, txn_id, schema.shape, value)
             .map_ok(Self::from)
             .await
     }
@@ -139,10 +150,60 @@ where
     where
         Shape: From<S>,
     {
-        BlockListFile::range(file, txn_id, shape.into(), start, stop)
+        let schema = Schema {
+            shape: shape.into(),
+            dtype: Ord::max(start.class(), stop.class()),
+        };
+
+        schema.validate("create Dense range")?;
+
+        BlockListFile::range(file, txn_id, schema.shape, start, stop)
             .map_ok(Self::from)
             .await
     }
+
+    pub async fn tile(
+        txn: T,
+        tensor: DenseTensor<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>,
+        multiples: Vec<u64>,
+    ) -> TCResult<Self> {
+        if multiples.len() != tensor.ndim() {
+            return Err(TCError::bad_request(
+                "wrong number of multiples to tile a Tensor with shape",
+                tensor.shape(),
+            ));
+        }
+
+        let txn_id = *txn.id();
+        let output_file = txn
+            .context()
+            .create_file_unique(txn_id, TensorType::Dense)
+            .await?;
+
+        let shape: Shape = tensor
+            .shape()
+            .iter()
+            .zip(&multiples)
+            .map(|(dim, m)| dim * m)
+            .collect();
+
+        let dtype = tensor.dtype();
+        let input = match tensor.blocks {
+            DenseAccessor::File(file) => DenseTensor::from(file),
+            other => {
+                let input_file = txn
+                    .context()
+                    .create_file_unique(txn_id, TensorType::Dense)
+                    .await?;
+
+                DenseTensor::copy_from(DenseTensor::from(other), input_file, &txn).await?
+            }
+        };
+
+        let output = Self::constant(output_file, txn_id, shape, dtype.zero()).await?;
+
+        tile(txn, input, output, multiples).await
+    }
 }
 
 impl<FD, FS, D, T> TensorPersist for DenseTensor<FD, FS, D, T, DenseAccessor<FD, FS, D, T>> {
@@ -498,6 +559,58 @@ where
     }
 }
 
+#[async_trait]
+impl<FD, FS, D, T, B> TensorDiagonal<D> for DenseTensor<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    B: DenseWrite<FD, FS, D, T>,
+    D::FileClass: From<TensorType>,
+{
+    type Txn = T;
+    type Diagonal = DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>;
+
+    async fn diagonal(self, txn: Self::Txn) -> TCResult<Self::Diagonal> {
+        if self.ndim() != 2 {
+            return Err(TCError::not_implemented(format!(
+                "diagonal of a {}-dimensional dense Tensor",
+                self.ndim()
+            )));
+        }
+
+        let size = self.shape()[0];
+        if size != self.shape()[1] {
+            return Err(TCError::bad_request(
+                "diagonal requires a square matrix but found",
+                self.shape(),
+            ));
+        }
+
+        let txn_id = *txn.id();
+        let file = txn
+            .context()
+            .create_file_unique(txn_id, TensorType::Dense)
+            .await?;
+
+        let dtype = self.dtype();
+        let blocks = self.blocks;
+
+        // TODO: is is really necessary to allocate a new Vec for every Coord?
+        let coords = futures::stream::iter((0..size).into_iter().map(|i| Ok(vec![i, i])));
+        let values = CoordBlocks::new(coords, 2, PER_BLOCK)
+            .map_ok(|coords| blocks.clone().read_values(txn.clone(), coords))
+            .try_buffered(num_cpus::get());
+
+        let shape = vec![size].into();
+        let blocks = BlockListFile::from_blocks(file, txn_id, Some(shape), dtype, values).await?;
+
+        Ok(blocks.into())
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T, B> TensorIO<D> for DenseTensor<FD, FS, D, T, B>
 where
@@ -532,7 +645,7 @@ where
 }
 
 #[async_trait]
-impl<FD, FS, D, T, B, O> TensorDualIO<D, DenseTensor<FD, FS, D, T, O>>
+impl<FD, FS, D, T, B> TensorDualIO<D, DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>>
     for DenseTensor<FD, FS, D, T, B>
 where
     D: Dir,
@@ -541,7 +654,6 @@ where
     FS: File<Node>,
     D::File: AsType<FD> + AsType<FS>,
     B: DenseWrite<FD, FS, D, T>,
-    O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
     type Txn = T;
@@ -550,13 +662,88 @@ where
         self,
         txn: T,
         bounds: Bounds,
-        other: DenseTensor<FD, FS, D, T, O>,
+        other: DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>,
     ) -> TCResult<()> {
         debug!("write {} to dense {}", other, bounds);
         self.blocks.write(txn, bounds, other.blocks).await
     }
 }
 
+#[async_trait]
+impl<FD, FS, D, T, B> TensorIndex<D> for DenseTensor<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    B: DenseWrite<FD, FS, D, T>,
+    D::FileClass: From<TensorType>,
+{
+    type Txn = T;
+    type Index = DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>;
+
+    async fn argmax(self, txn: Self::Txn, axis: usize) -> TCResult<Self::Index> {
+        if axis >= self.ndim() {
+            return Err(TCError::unsupported(format!(
+                "invalid argmax axis for tensor with {} dimensions: {}",
+                self.ndim(),
+                axis
+            )));
+        }
+
+        let dtype = NumberType::UInt(UIntType::U64);
+        let shape = {
+            let mut shape = self.shape().clone();
+            shape.remove(axis);
+            shape
+        };
+
+        let txn_id = *txn.id();
+        let file = txn
+            .context()
+            .create_file_unique(txn_id, TensorType::Dense)
+            .await?;
+
+        let per_block = self.size() / shape.size();
+        debug_assert_eq!(self.size() / per_block, shape.size());
+
+        let blocks = self.blocks.value_stream(txn).await?; // TODO: just use self.block_stream directly
+        let values = blocks
+            .chunks(per_block as usize)
+            .map(|values| values.into_iter().collect::<TCResult<Vec<Number>>>())
+            .map_ok(Array::from)
+            .map_ok(|array| {
+                debug_assert!(array.len() as u64 == per_block);
+                let (i, _max) = array.argmax();
+                Number::from(i as u64)
+            });
+
+        BlockListFile::from_values(file, txn_id, shape, dtype, values)
+            .map_ok(DenseTensor::from)
+            .await
+    }
+
+    async fn argmax_all(self, txn: Self::Txn) -> TCResult<u64> {
+        let mut offset = 0;
+        let mut max_value = self.dtype().zero();
+        let mut argmax = 0;
+
+        let mut blocks = self.blocks.block_stream(txn).await?;
+        while let Some(block) = blocks.try_next().await? {
+            let (i, max) = block.argmax();
+            if max > max_value {
+                argmax = offset + (i as u64);
+                max_value = max;
+            }
+
+            offset += block.len() as u64;
+        }
+
+        Ok(argmax)
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T, B> TensorDualIO<D, Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
 where
@@ -581,8 +768,27 @@ where
         };
 
         match other {
-            Tensor::Dense(dense) => self.write(txn, bounds, dense).await,
-            Tensor::Sparse(sparse) => self.write(txn, bounds, sparse.into_dense()).await,
+            Tensor::Dense(dense) => match dense.blocks {
+                DenseAccessor::File(file) => self.write(txn, bounds, DenseTensor::from(file)).await,
+                other => {
+                    let file = txn
+                        .context()
+                        .create_file_unique(*txn.id(), TensorType::Dense)
+                        .await?;
+
+                    let other = DenseTensor::copy_from(other.into(), file, &txn).await?;
+                    self.write(txn, bounds, other).await
+                }
+            },
+            Tensor::Sparse(sparse) => {
+                let file = txn
+                    .context()
+                    .create_file_unique(*txn.id(), TensorType::Dense)
+                    .await?;
+
+                let other = DenseTensor::copy_from(sparse.into_dense(), file, &txn).await?;
+                self.write(txn, bounds, other).await
+            }
         }
     }
 }
@@ -626,6 +832,15 @@ where
         self.combine(other, div_array, Div::div, dtype)
     }
 
+    fn log(self, base: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::LeftCombine> {
+        fn log(n: Number, base: Number) -> Number {
+            n.log(Float::cast_from(base))
+        }
+
+        let dtype = self.dtype().one().ln().class();
+        self.combine(base, Array::log, log, dtype)
+    }
+
     fn mul(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn mul_array(l: &Array, r: &Array) -> Array {
             debug_assert_eq!(l.len(), r.len());
@@ -688,6 +903,13 @@ where
         }
     }
 
+    fn log(self, base: Tensor<FD, FS, D, T>) -> TCResult<Self::LeftCombine> {
+        match base {
+            Tensor::Dense(dense) => self.log(dense).map(Tensor::from),
+            Tensor::Sparse(sparse) => self.into_sparse().log(sparse).map(Tensor::from),
+        }
+    }
+
     fn mul(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
             Tensor::Dense(dense) => self.mul(dense).map(Tensor::from),
@@ -736,6 +958,28 @@ where
         Ok(BlockListConst::new(self.blocks, other, div_array, Number::div).into())
     }
 
+    fn log_const(self, base: Number) -> TCResult<Self::Combine> {
+        if base.class().is_complex() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
+
+        let base = Number::Float(base.cast_into());
+
+        fn log(n: Number, base: Number) -> Number {
+            if let Number::Float(base) = base {
+                n.log(base)
+            } else {
+                unreachable!("log with non-floating point base")
+            }
+        }
+
+        fn log_array(l: Array, r: Number) -> Array {
+            l.log_const(r)
+        }
+
+        Ok(BlockListConst::new(self.blocks, base, log_array, log).into())
+    }
+
     fn mul_const(self, other: Number) -> TCResult<Self::Combine> {
         fn mul_array(l: Array, r: Number) -> Array {
             &l * r
@@ -849,6 +1093,8 @@ where
     type Broadcast = DenseTensor<FD, FS, D, T, BlockListBroadcast<FD, FS, D, T, B>>;
     type Cast = DenseTensor<FD, FS, D, T, BlockListCast<FD, FS, D, T, B>>;
     type Expand = DenseTensor<FD, FS, D, T, BlockListExpand<FD, FS, D, T, B>>;
+    type Flip = DenseTensor<FD, FS, D, T, BlockListFlip<FD, FS, D, T, B>>;
+    type Reshape = DenseTensor<FD, FS, D, T, BlockListReshape<FD, FS, D, T, B>>;
     type Slice = DenseTensor<FD, FS, D, T, B::Slice>;
     type Transpose = DenseTensor<FD, FS, D, T, B::Transpose>;
 
@@ -867,6 +1113,16 @@ where
         Ok(DenseTensor::from(blocks))
     }
 
+    fn flip(self, axis: usize) -> TCResult<Self::Flip> {
+        let blocks = BlockListFlip::new(self.blocks, axis)?;
+        Ok(DenseTensor::from(blocks))
+    }
+
+    fn reshape(self, shape: Shape) -> TCResult<Self::Reshape> {
+        let blocks = BlockListReshape::new(self.blocks, shape)?;
+        Ok(DenseTensor::from(blocks))
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         let blocks = self.blocks.slice(bounds)?;
         Ok(DenseTensor::from(blocks))
@@ -878,6 +1134,43 @@ where
     }
 }
 
+macro_rules! trig {
+    ($fun:ident) => {
+        fn $fun(&self) -> TCResult<Self::Unary> {
+            let dtype = trig_dtype(self.dtype());
+            let blocks = BlockListUnary::new(self.blocks.clone(), Array::$fun, Number::$fun, dtype);
+            Ok(DenseTensor::from(blocks))
+        }
+    };
+}
+
+#[async_trait]
+impl<FD, FS, D, T, B> TensorTrig for DenseTensor<FD, FS, D, T, B>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    B: DenseAccess<FD, FS, D, T>,
+{
+    type Unary = DenseTensor<FD, FS, D, T, BlockListUnary<FD, FS, D, T, B>>;
+
+    trig! {asin}
+    trig! {sin}
+    trig! {sinh}
+    trig! {asinh}
+
+    trig! {acos}
+    trig! {cos}
+    trig! {cosh}
+    trig! {acosh}
+
+    trig! {atan}
+    trig! {tan}
+    trig! {tanh}
+    trig! {atanh}
+}
+
 #[async_trait]
 impl<FD, FS, D, T, B> TensorUnary<D> for DenseTensor<FD, FS, D, T, B>
 where
@@ -897,7 +1190,7 @@ where
             self.blocks.clone(),
             Array::abs,
             <Number as NumberInstance>::abs,
-            NumberType::Bool,
+            self.dtype().one().abs().class(),
         );
 
         Ok(DenseTensor::from(blocks))
@@ -919,6 +1212,18 @@ where
         Ok(DenseTensor::from(blocks))
     }
 
+    fn ln(&self) -> TCResult<Self::Unary> {
+        let dtype = self.dtype().one().ln().class();
+        let blocks = BlockListUnary::new(self.blocks.clone(), Array::ln, Number::ln, dtype);
+        Ok(DenseTensor::from(blocks))
+    }
+
+    fn round(&self) -> TCResult<Self::Unary> {
+        let dtype = self.dtype().one().round().class();
+        let blocks = BlockListUnary::new(self.blocks.clone(), Array::round, Number::round, dtype);
+        Ok(DenseTensor::from(blocks))
+    }
+
     async fn all(self, txn: T) -> TCResult<bool> {
         let mut blocks = self.blocks.block_stream(txn).await?;
 
@@ -974,25 +1279,6 @@ where
     }
 }
 
-#[async_trait]
-impl<'en, FD, FS, D, T, B> Hash<'en, D> for DenseTensor<FD, FS, D, T, B>
-where
-    D: Dir,
-    T: Transaction<D>,
-    FD: File<Array>,
-    FS: File<Node>,
-    D::File: AsType<FD> + AsType<FS>,
-    B: DenseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
-{
-    type Item = Array;
-    type Txn = T;
-
-    async fn hashable(&'en self, txn: &'en T) -> TCResult<TCBoxTryStream<'en, Self::Item>> {
-        self.blocks.clone().block_stream(txn.clone()).await
-    }
-}
-
 #[async_trait]
 impl<FD, FS, D, T, B> CopyFrom<D, DenseTensor<FD, FS, D, T, B>>
     for DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>
@@ -1147,8 +1433,9 @@ where
     async fn visit_seq<A: de::SeqAccess>(self, mut seq: A) -> Result<Self::Value, A::Error> {
         debug!("deserialize DenseTensor");
 
-        let schema = seq.next_element(()).await?;
+        let schema = seq.next_element::<Schema>(()).await?;
         let schema = schema.ok_or_else(|| de::Error::invalid_length(0, "a tensor schema"))?;
+        schema.validate("load Dense").map_err(de::Error::custom)?;
         debug!("DenseTensor schema is {}", schema);
 
         let cxt = (self.txn_id, self.file, schema);
@@ -1303,3 +1590,17 @@ fn encodable_c64<'en>(blocks: TCBoxTryStream<'en, Array>) -> impl Stream<Item =
             dest
         })
 }
+
+#[inline]
+fn array_err(err: afarray::ArrayError) -> TCError {
+    TCError::new(ErrorType::BadRequest, err.to_string())
+}
+
+#[inline]
+fn div_ceil(l: u64, r: u64) -> u64 {
+    if l % r == 0 {
+        l / r
+    } else {
+        (l / r) + 1
+    }
+}
