@@ -4,7 +4,9 @@ use std::marker::PhantomData;
 
 use async_trait::async_trait;
 use destream::{de, en};
-use futures::TryFutureExt;
+use futures::future::TryFutureExt;
+use futures::stream::{StreamExt, TryStreamExt};
+use itertools::Itertools;
 use log::debug;
 use safecast::*;
 
@@ -12,21 +14,21 @@ use tc_btree::{BTreeType, Node};
 use tc_error::*;
 use tc_transact::fs::{Dir, File};
 use tc_transact::{IntoView, Transaction, TxnId};
-use tc_value::{Number, NumberType, Value, ValueType};
+use tc_value::{FloatType, IntType, Number, NumberType, UIntType, Value, ValueType};
 use tcgeneric::{
     label, path_label, Class, Instance, NativeClass, PathLabel, PathSegment, TCBoxTryFuture,
     TCPathBuf, Tuple,
 };
 
+use stream::ReadValueAt;
+
 pub use afarray::{print_af_info, Array};
 pub use bounds::{AxisBounds, Bounds, Shape};
-pub use dense::{BlockListFile, DenseAccess, DenseAccessor, DenseTensor, DenseWrite};
-pub use einsum::einsum;
+pub use dense::{arg_sort, BlockListFile, DenseAccess, DenseAccessor, DenseTensor, DenseWrite};
 pub use sparse::{SparseAccess, SparseAccessor, SparseTable, SparseTensor, SparseWrite};
 
 mod bounds;
 mod dense;
-mod einsum;
 mod sparse;
 mod stream;
 mod transform;
@@ -47,6 +49,15 @@ pub struct Schema {
     pub dtype: NumberType,
 }
 
+impl Schema {
+    /// Raise an error if this schema is invalid.
+    ///
+    /// `debug_info` should be the name of the calling Op, as a user would understand it.
+    pub fn validate(&self, debug_info: &'static str) -> TCResult<()> {
+        self.shape.validate(debug_info)
+    }
+}
+
 impl TryCastFrom<Value> for Schema {
     fn can_cast_from(value: &Value) -> bool {
         match value {
@@ -223,6 +234,18 @@ pub trait TensorCompareConst {
     fn ne_const(self, other: Number) -> TCResult<Self::Compare>;
 }
 
+/// [`Tensor`] linear algebra operations
+#[async_trait]
+pub trait TensorDiagonal<D: Dir> {
+    /// The type of [`Transaction`] to expect
+    type Txn: Transaction<D>;
+
+    /// The type of [`Tensor`] returned by `diagonal`
+    type Diagonal: TensorAccess;
+
+    async fn diagonal(self, txn: Self::Txn) -> TCResult<Self::Diagonal>;
+}
+
 /// [`Tensor`] I/O operations
 #[async_trait]
 pub trait TensorIO<D: Dir> {
@@ -249,6 +272,22 @@ pub trait TensorDualIO<D: Dir, O> {
     async fn write(self, txn: Self::Txn, bounds: Bounds, value: O) -> TCResult<()>;
 }
 
+/// [`Tensor`] indexing operations
+#[async_trait]
+pub trait TensorIndex<D: Dir> {
+    /// The type of [`Transaction`] to expect
+    type Txn: Transaction<D>;
+
+    /// The type of [`Tensor`] returned by `argmax`.
+    type Index: TensorInstance;
+
+    /// Return the indices of the maximum values in this [`Tensor`] along the given `axis`.
+    async fn argmax(self, txn: Self::Txn, axis: usize) -> TCResult<Self::Index>;
+
+    /// Return the offset of the maximum value in this [`Tensor`].
+    async fn argmax_all(self, txn: Self::Txn) -> TCResult<u64>;
+}
+
 /// [`Tensor`] math operations
 pub trait TensorMath<D: Dir, O> {
     /// The result type of a math operation
@@ -263,10 +302,13 @@ pub trait TensorMath<D: Dir, O> {
     /// Divide `self` by `other`.
     fn div(self, other: O) -> TCResult<Self::LeftCombine>;
 
+    /// Element-wise logarithm of `self` with respect to the given `base`.
+    fn log(self, base: O) -> TCResult<Self::LeftCombine>;
+
     /// Multiply two tensors together.
     fn mul(self, other: O) -> TCResult<Self::LeftCombine>;
 
-    /// Raise `self` to the power `other`.
+    /// Raise `self` to the power of `other`.
     fn pow(self, other: O) -> TCResult<Self::LeftCombine>;
 
     /// Subtract `other` from `self`.
@@ -284,6 +326,9 @@ pub trait TensorMathConst {
     /// Divide `self` by `other`.
     fn div_const(self, other: Number) -> TCResult<Self::Combine>;
 
+    /// Element-wise logarithm
+    fn log_const(self, base: Number) -> TCResult<Self::Combine>;
+
     /// Multiply `self` by `other`.
     fn mul_const(self, other: Number) -> TCResult<Self::Combine>;
 
@@ -335,6 +380,12 @@ pub trait TensorTransform {
     /// A [`Tensor`] with an expanded dimension
     type Expand: TensorInstance;
 
+    /// A [`Tensor`] flipped around one axis
+    type Flip: TensorInstance;
+
+    /// A reshaped [`Tensor`]
+    type Reshape: TensorInstance;
+
     /// A [`Tensor`] slice
     type Slice: TensorInstance;
 
@@ -350,6 +401,12 @@ pub trait TensorTransform {
     /// Insert a new dimension of size 1 at the given `axis`.
     fn expand_dims(self, axis: usize) -> TCResult<Self::Expand>;
 
+    /// Flip this [`Tensor`] around the given `axis`.
+    fn flip(self, axis: usize) -> TCResult<Self::Flip>;
+
+    /// Change the shape of this [`Tensor`].
+    fn reshape(self, shape: Shape) -> TCResult<Self::Reshape>;
+
     /// Return a slice of this [`Tensor`] with the given `bounds`.
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice>;
 
@@ -370,9 +427,15 @@ pub trait TensorUnary<D: Dir> {
     /// Element-wise absolute value
     fn abs(&self) -> TCResult<Self::Unary>;
 
-    /// Raise `e` to the power of `self`
+    /// Element-wise exponentiation
     fn exp(&self) -> TCResult<Self::Unary>;
 
+    /// Element-wise natural logarithm
+    fn ln(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise round to the nearest integer
+    fn round(&self) -> TCResult<Self::Unary>;
+
     /// Return `true` if all elements in this [`Tensor`] are nonzero.
     async fn all(self, txn: Self::Txn) -> TCResult<bool>;
 
@@ -383,6 +446,63 @@ pub trait TensorUnary<D: Dir> {
     fn not(&self) -> TCResult<Self::Unary>;
 }
 
+/// Trigonometric [`Tensor`] operations
+#[async_trait]
+pub trait TensorTrig {
+    /// The return type of a unary operation
+    type Unary: TensorInstance;
+
+    /// Element-wise arcsine
+    fn asin(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise sine
+    fn sin(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise hyperbolic arcsine
+    fn asinh(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise hyperbolic sine
+    fn sinh(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise arccosine
+    fn acos(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise cosine
+    fn cos(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise hyperbolic arccosine
+    fn acosh(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise hyperbolic cosine
+    fn cosh(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise arctangent
+    fn atan(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise tangent
+    fn tan(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise hyperbolic tangent
+    fn tanh(&self) -> TCResult<Self::Unary>;
+
+    /// Element-wise hyperbolic arctangent
+    fn atanh(&self) -> TCResult<Self::Unary>;
+}
+
+fn trig_dtype(dtype: NumberType) -> NumberType {
+    match dtype {
+        NumberType::Int(it) => match it {
+            IntType::I64 => FloatType::F64.into(),
+            _ => FloatType::F32.into(),
+        },
+        NumberType::UInt(ut) => match ut {
+            UIntType::U64 => FloatType::F64.into(),
+            _ => FloatType::F32.into(),
+        },
+        other => other,
+    }
+}
+
 /// The [`Class`] of [`Tensor`]
 #[derive(Clone, Copy, Eq, PartialEq)]
 pub enum TensorType {
@@ -435,6 +555,7 @@ where
     D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
 {
+    /// Get the [`Schema`] of this [`Tensor`]
     pub fn schema(&self) -> Schema {
         Schema {
             dtype: self.dtype(),
@@ -695,6 +816,28 @@ where
     }
 }
 
+#[async_trait]
+impl<FD, FS, D, T> TensorDiagonal<D> for Tensor<FD, FS, D, T>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
+    SparseTable<FD, FS, D, T>: ReadValueAt<D, Txn = T>,
+{
+    type Txn = T;
+    type Diagonal = Self;
+
+    async fn diagonal(self, txn: Self::Txn) -> TCResult<Self::Diagonal> {
+        match self {
+            Self::Dense(dense) => dense.diagonal(txn).map_ok(Self::from).await,
+            Self::Sparse(sparse) => sparse.diagonal(txn).map_ok(Self::from).await,
+        }
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T> TensorIO<D> for Tensor<FD, FS, D, T>
 where
@@ -739,11 +882,11 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D::File: AsType<FD> + AsType<FS>,
-    D::FileClass: From<TensorType>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
 {
     type Txn = T;
 
-    async fn write(self, txn: T, bounds: Bounds, value: Self) -> TCResult<()> {
+    async fn write(self, txn: T, bounds: Bounds, value: Tensor<FD, FS, D, T>) -> TCResult<()> {
         debug!("Tensor::write {} to {}", value, bounds);
 
         match self {
@@ -753,6 +896,34 @@ where
     }
 }
 
+#[async_trait]
+impl<FD, FS, D, T> TensorIndex<D> for Tensor<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
+{
+    type Txn = T;
+    type Index = Self;
+
+    async fn argmax(self, txn: Self::Txn, axis: usize) -> TCResult<Self::Index> {
+        match self {
+            Self::Dense(dense) => dense.argmax(txn, axis).map_ok(Self::from).await,
+            Self::Sparse(sparse) => sparse.argmax(txn, axis).map_ok(Self::from).await,
+        }
+    }
+
+    async fn argmax_all(self, txn: Self::Txn) -> TCResult<u64> {
+        match self {
+            Self::Dense(dense) => dense.argmax_all(txn).await,
+            Self::Sparse(sparse) => sparse.argmax_all(txn).await,
+        }
+    }
+}
+
 impl<FD, FS, D, T> TensorMath<D, Self> for Tensor<FD, FS, D, T>
 where
     D: Dir,
@@ -779,6 +950,13 @@ where
         }
     }
 
+    fn log(self, base: Self) -> TCResult<Self::LeftCombine> {
+        match self {
+            Self::Dense(this) => this.log(base),
+            Self::Sparse(this) => this.log(base),
+        }
+    }
+
     fn mul(self, other: Self) -> TCResult<Self::LeftCombine> {
         match self {
             Self::Dense(this) => this.mul(other),
@@ -826,6 +1004,13 @@ where
         }
     }
 
+    fn log_const(self, base: Number) -> TCResult<Self::Combine> {
+        match self {
+            Self::Dense(dense) => dense.log_const(base).map(Self::from),
+            Self::Sparse(sparse) => sparse.log_const(base).map(Self::from),
+        }
+    }
+
     fn mul_const(self, other: Number) -> TCResult<Self::Combine> {
         match self {
             Self::Dense(dense) => dense.mul_const(other).map(Self::from),
@@ -901,6 +1086,8 @@ where
     type Broadcast = Self;
     type Cast = Self;
     type Expand = Self;
+    type Flip = Self;
+    type Reshape = Self;
     type Slice = Self;
     type Transpose = Self;
 
@@ -933,6 +1120,20 @@ where
         }
     }
 
+    fn flip(self, axis: usize) -> TCResult<Self::Flip> {
+        match self {
+            Self::Dense(dense) => dense.flip(axis).map(Self::from),
+            Self::Sparse(sparse) => sparse.flip(axis).map(Self::from),
+        }
+    }
+
+    fn reshape(self, shape: Shape) -> TCResult<Self::Reshape> {
+        match self {
+            Self::Dense(dense) => dense.reshape(shape).map(Self::from),
+            Self::Sparse(sparse) => sparse.reshape(shape).map(Self::from),
+        }
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self> {
         if bounds == Bounds::all(self.shape()) {
             return Ok(self);
@@ -945,8 +1146,10 @@ where
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self> {
-        if permutation == Some((0..self.ndim()).collect()) {
-            return Ok(self);
+        if let Some(permutation) = &permutation {
+            if permutation.iter().zip(0..self.ndim()).all(|(x, i)| x == &i) {
+                return Ok(self);
+            }
         }
 
         match self {
@@ -956,6 +1159,44 @@ where
     }
 }
 
+macro_rules! trig {
+    ($fun:ident) => {
+        fn $fun(&self) -> TCResult<Self> {
+            match self {
+                Self::Dense(dense) => dense.$fun().map(Self::from),
+                Self::Sparse(sparse) => sparse.$fun().map(Self::from),
+            }
+        }
+    };
+}
+
+impl<FD, FS, D, T> TensorTrig for Tensor<FD, FS, D, T>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+{
+    type Unary = Self;
+
+    trig! {asin}
+    trig! {sin}
+    trig! {asinh}
+    trig! {sinh}
+
+    trig! {acos}
+    trig! {cos}
+    trig! {acosh}
+    trig! {cosh}
+
+    trig! {atan}
+    trig! {tan}
+    trig! {atanh}
+    trig! {tanh}
+}
+
 #[async_trait]
 impl<FD, FS, D, T> TensorUnary<D> for Tensor<FD, FS, D, T>
 where
@@ -969,7 +1210,7 @@ where
     type Txn = T;
     type Unary = Self;
 
-    fn abs(&self) -> TCResult<Self> {
+    fn abs(&self) -> TCResult<Self::Unary> {
         match self {
             Self::Dense(dense) => dense.abs().map(Self::from),
             Self::Sparse(sparse) => sparse.abs().map(Self::from),
@@ -983,6 +1224,20 @@ where
         }
     }
 
+    fn ln(&self) -> TCResult<Self::Unary> {
+        match self {
+            Self::Dense(dense) => dense.ln().map(Self::from),
+            Self::Sparse(sparse) => sparse.ln().map(Self::from),
+        }
+    }
+
+    fn round(&self) -> TCResult<Self::Unary> {
+        match self {
+            Self::Dense(dense) => dense.round().map(Self::from),
+            Self::Sparse(sparse) => sparse.round().map(Self::from),
+        }
+    }
+
     async fn all(self, txn: T) -> TCResult<bool> {
         match self {
             Self::Dense(dense) => dense.all(txn).await,
@@ -1148,17 +1403,33 @@ impl<'en> en::IntoStream<'en> for TensorView<'en> {
     }
 }
 
-impl<FD, FS, D, T> fmt::Debug for Tensor<FD, FS, D, T> {
+impl<FD, FS, D, T> fmt::Debug for Tensor<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+{
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         fmt::Display::fmt(self, f)
     }
 }
 
-impl<FD, FS, D, T> fmt::Display for Tensor<FD, FS, D, T> {
+impl<FD, FS, D, T> fmt::Display for Tensor<FD, FS, D, T>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+{
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            Self::Sparse(_) => f.write_str("a sparse tensor"),
-            Self::Dense(_) => f.write_str("a dense tensor"),
+            Self::Dense(dense) => fmt::Display::fmt(dense, f),
+            Self::Sparse(sparse) => fmt::Display::fmt(sparse, f),
         }
     }
 }
@@ -1234,6 +1505,47 @@ impl<FD, FS, D, T> Default for Phantom<FD, FS, D, T> {
     }
 }
 
+async fn tile<D, T, I, O>(txn: T, input: I, output: O, multiples: Vec<u64>) -> TCResult<O>
+where
+    D: Dir,
+    T: Transaction<D>,
+    I: TensorAccess + Clone,
+    O: TensorAccess + TensorDualIO<D, I, Txn = T> + Clone,
+{
+    let max_usize = usize::MAX as u64;
+    let input_shape = input.shape();
+
+    if input_shape.iter().any(|dim| dim >= &max_usize) {
+        return Err(TCError::unsupported(format!(
+            "Tensor of shape {} is too large to tile",
+            input_shape
+        )));
+    }
+
+    let slices = input_shape
+        .iter()
+        .zip(&multiples)
+        .map(|(dim, m)| (0..(dim * m)).step_by(*dim as usize));
+
+    let mut writes = Vec::with_capacity((output.size() / input.size()) as usize);
+    for slice in Itertools::multi_cartesian_product(slices) {
+        let bounds = slice
+            .into_iter()
+            .zip(input_shape.iter())
+            .map(|(i, dim)| AxisBounds::In(i..(i + *dim)))
+            .collect();
+
+        writes.push(output.clone().write(txn.clone(), bounds, input.clone()));
+    }
+
+    futures::stream::iter(writes)
+        .buffer_unordered(num_cpus::get())
+        .try_fold((), |(), ()| futures::future::ready(Ok(())))
+        .await?;
+
+    Ok(output)
+}
+
 #[inline]
 fn coord_bounds(shape: &[u64]) -> Vec<u64> {
     (0..shape.len())
