@@ -13,7 +13,7 @@ use tc_error::*;
 use tc_transact::fs::{Dir, File};
 use tc_transact::{Transaction, TxnId};
 use tc_value::{FloatInstance, Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
+use tcgeneric::{TCBoxTryFuture, TCBoxTryStream, Tuple};
 
 use crate::dense::{DenseAccess, DenseAccessor, DenseTensor, PER_BLOCK};
 use crate::stream::{sorted_coords, sorted_values, Read, ReadValueAt};
@@ -29,7 +29,7 @@ use super::{SparseRow, SparseStream, SparseTensor};
 /// Access methods for [`SparseTensor`] data
 #[async_trait]
 pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
-    Clone + fmt::Display + ReadValueAt<D, Txn = T> + TensorAccess + Send + Sync + 'static
+    ReadValueAt<D, Txn = T> + TensorAccess + Clone + fmt::Display + Send + Sync + 'static
 {
     /// The type of a slice of this accessor
     type Slice: SparseAccess<FD, FS, D, T>;
@@ -82,13 +82,36 @@ pub enum SparseAccessor<FD, FS, D, T> {
     CombineLeft(Box<SparseLeftCombinator<FD, FS, D, T, Self, Self>>),
     Dense(Box<DenseToSparse<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>>),
     Expand(Box<SparseExpand<FD, FS, D, T, Self>>),
+    Flip(Box<SparseFlip<FD, FS, D, T, Self>>),
     Slice(SparseTableSlice<FD, FS, D, T>),
     Reduce(Box<SparseReduce<FD, FS, D, T>>),
+    Reshape(Box<SparseReshape<FD, FS, D, T, Self>>),
     Table(SparseTable<FD, FS, D, T>),
     Transpose(Box<SparseTranspose<FD, FS, D, T, Self>>),
     Unary(Box<SparseUnary<FD, FS, D, T>>),
 }
 
+macro_rules! dispatch {
+    ($this:ident, $var:ident, $call:expr) => {
+        match $this {
+            Self::Broadcast($var) => $call,
+            Self::Cast($var) => $call,
+            Self::Combine($var) => $call,
+            Self::CombineConst($var) => $call,
+            Self::CombineLeft($var) => $call,
+            Self::Dense($var) => $call,
+            Self::Expand($var) => $call,
+            Self::Flip($var) => $call,
+            Self::Slice($var) => $call,
+            Self::Reduce($var) => $call,
+            Self::Reshape($var) => $call,
+            Self::Table($var) => $call,
+            Self::Transpose($var) => $call,
+            Self::Unary($var) => $call,
+        }
+    };
+}
+
 impl<FD, FS, D, T> TensorAccess for SparseAccessor<FD, FS, D, T>
 where
     D: Dir,
@@ -99,71 +122,19 @@ where
     D::FileClass: From<TensorType>,
 {
     fn dtype(&self) -> NumberType {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.dtype(),
-            Self::Cast(cast) => cast.dtype(),
-            Self::Combine(combine) => combine.dtype(),
-            Self::CombineConst(combine) => combine.dtype(),
-            Self::CombineLeft(combine) => combine.dtype(),
-            Self::Dense(dense) => dense.dtype(),
-            Self::Expand(expand) => expand.dtype(),
-            Self::Slice(slice) => slice.dtype(),
-            Self::Reduce(reduce) => reduce.dtype(),
-            Self::Table(table) => table.dtype(),
-            Self::Transpose(transpose) => transpose.dtype(),
-            Self::Unary(unary) => unary.dtype(),
-        }
+        dispatch!(self, this, this.dtype())
     }
 
     fn ndim(&self) -> usize {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.ndim(),
-            Self::Cast(cast) => cast.ndim(),
-            Self::Combine(combine) => combine.ndim(),
-            Self::CombineConst(combine) => combine.ndim(),
-            Self::CombineLeft(combine) => combine.ndim(),
-            Self::Dense(dense) => dense.ndim(),
-            Self::Expand(expand) => expand.ndim(),
-            Self::Slice(slice) => slice.ndim(),
-            Self::Reduce(reduce) => reduce.ndim(),
-            Self::Table(table) => table.ndim(),
-            Self::Transpose(transpose) => transpose.ndim(),
-            Self::Unary(unary) => unary.ndim(),
-        }
+        dispatch!(self, this, this.ndim())
     }
 
     fn shape(&self) -> &Shape {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.shape(),
-            Self::Cast(cast) => cast.shape(),
-            Self::Combine(combine) => combine.shape(),
-            Self::CombineConst(combine) => combine.shape(),
-            Self::CombineLeft(combine) => combine.shape(),
-            Self::Dense(dense) => dense.shape(),
-            Self::Expand(expand) => expand.shape(),
-            Self::Reduce(reduce) => reduce.shape(),
-            Self::Slice(slice) => slice.shape(),
-            Self::Table(table) => table.shape(),
-            Self::Transpose(transpose) => transpose.shape(),
-            Self::Unary(unary) => unary.shape(),
-        }
+        dispatch!(self, this, this.shape())
     }
 
     fn size(&self) -> u64 {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.size(),
-            Self::Cast(cast) => cast.size(),
-            Self::Combine(combine) => combine.size(),
-            Self::CombineConst(combine) => combine.size(),
-            Self::CombineLeft(combine) => combine.size(),
-            Self::Dense(dense) => dense.size(),
-            Self::Expand(expand) => expand.size(),
-            Self::Slice(slice) => slice.size(),
-            Self::Reduce(reduce) => reduce.size(),
-            Self::Table(table) => table.size(),
-            Self::Transpose(transpose) => transpose.size(),
-            Self::Unary(unary) => unary.size(),
-        }
+        dispatch!(self, this, this.size())
     }
 }
 
@@ -185,109 +156,27 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.filled(txn).await,
-            Self::Cast(cast) => cast.filled(txn).await,
-            Self::Combine(combine) => combine.filled(txn).await,
-            Self::CombineConst(combine) => combine.filled(txn).await,
-            Self::CombineLeft(combine) => combine.filled(txn).await,
-            Self::Dense(dense) => dense.filled(txn).await,
-            Self::Expand(expand) => expand.filled(txn).await,
-            Self::Reduce(reduce) => reduce.filled(txn).await,
-            Self::Slice(slice) => slice.filled(txn).await,
-            Self::Table(table) => table.filled(txn).await,
-            Self::Transpose(transpose) => transpose.filled(txn).await,
-            Self::Unary(unary) => unary.filled(txn).await,
-        }
+        dispatch!(self, this, this.filled(txn).await)
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.filled_at(txn, axes).await,
-            Self::Cast(cast) => cast.filled_at(txn, axes).await,
-            Self::Combine(combine) => combine.filled_at(txn, axes).await,
-            Self::CombineConst(combine) => combine.filled_at(txn, axes).await,
-            Self::CombineLeft(combine) => combine.filled_at(txn, axes).await,
-            Self::Dense(dense) => dense.filled_at(txn, axes).await,
-            Self::Expand(expand) => expand.filled_at(txn, axes).await,
-            Self::Reduce(reduce) => reduce.filled_at(txn, axes).await,
-            Self::Slice(slice) => slice.filled_at(txn, axes).await,
-            Self::Table(table) => table.filled_at(txn, axes).await,
-            Self::Transpose(transpose) => transpose.filled_at(txn, axes).await,
-            Self::Unary(unary) => unary.filled_at(txn, axes).await,
-        }
+        dispatch!(self, this, this.filled_at(txn, axes).await)
     }
 
     async fn filled_count(self, txn: T) -> TCResult<u64> {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.filled_count(txn).await,
-            Self::Cast(cast) => cast.filled_count(txn).await,
-            Self::Combine(combine) => combine.filled_count(txn).await,
-            Self::CombineConst(combine) => combine.filled_count(txn).await,
-            Self::CombineLeft(combine) => combine.filled_count(txn).await,
-            Self::Dense(dense) => dense.filled_count(txn).await,
-            Self::Expand(expand) => expand.filled_count(txn).await,
-            Self::Reduce(reduce) => reduce.filled_count(txn).await,
-            Self::Slice(slice) => slice.filled_count(txn).await,
-            Self::Table(table) => table.filled_count(txn).await,
-            Self::Transpose(transpose) => transpose.filled_count(txn).await,
-            Self::Unary(unary) => unary.filled_count(txn).await,
-        }
+        dispatch!(self, this, this.filled_count(txn).await)
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self> {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.slice(bounds).map(SparseAccess::accessor),
-            Self::Cast(cast) => cast.slice(bounds).map(SparseAccess::accessor),
-            Self::Combine(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
-            Self::CombineConst(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
-            Self::CombineLeft(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
-            Self::Dense(dense) => dense.slice(bounds).map(SparseAccess::accessor),
-            Self::Expand(expand) => expand.slice(bounds).map(SparseAccess::accessor),
-            Self::Reduce(reduce) => reduce.slice(bounds).map(SparseAccess::accessor),
-            Self::Slice(slice) => slice.slice(bounds).map(SparseAccess::accessor),
-            Self::Table(table) => table.slice(bounds).map(SparseAccess::accessor),
-            Self::Transpose(transpose) => transpose.slice(bounds).map(SparseAccess::accessor),
-            Self::Unary(unary) => unary.slice(bounds).map(SparseAccess::accessor),
-        }
+        dispatch!(self, this, this.slice(bounds).map(SparseAccess::accessor))
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self> {
-        match self {
-            Self::Broadcast(broadcast) => {
-                broadcast.transpose(permutation).map(SparseAccess::accessor)
-            }
-
-            Self::Cast(cast) => cast.transpose(permutation).map(SparseAccess::accessor),
-
-            Self::Combine(combinator) => combinator
-                .transpose(permutation)
-                .map(SparseAccess::accessor),
-
-            Self::CombineConst(combinator) => combinator
-                .transpose(permutation)
-                .map(SparseAccess::accessor),
-
-            Self::CombineLeft(combinator) => combinator
-                .transpose(permutation)
-                .map(SparseAccess::accessor),
-
-            Self::Dense(dense) => dense.transpose(permutation).map(SparseAccess::accessor),
-
-            Self::Expand(expand) => expand.transpose(permutation).map(SparseAccess::accessor),
-
-            Self::Reduce(reduce) => reduce.transpose(permutation).map(SparseAccess::accessor),
-
-            Self::Table(table) => table.transpose(permutation).map(SparseAccess::accessor),
-
-            Self::Slice(slice) => slice.transpose(permutation).map(SparseAccess::accessor),
-
-            Self::Transpose(transpose) => {
-                transpose.transpose(permutation).map(SparseAccess::accessor)
-            }
-
-            Self::Unary(unary) => unary.transpose(permutation).map(SparseAccess::accessor),
-        }
+        dispatch!(
+            self,
+            this,
+            this.transpose(permutation).map(SparseAccess::accessor)
+        )
     }
 }
 
@@ -321,39 +210,13 @@ where
     type Txn = T;
 
     fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.read_value_at(txn, coord),
-            Self::Cast(cast) => cast.read_value_at(txn, coord),
-            Self::Combine(combine) => combine.read_value_at(txn, coord),
-            Self::CombineConst(combine) => combine.read_value_at(txn, coord),
-            Self::CombineLeft(combine) => combine.read_value_at(txn, coord),
-            Self::Dense(dense) => dense.read_value_at(txn, coord),
-            Self::Expand(expand) => expand.read_value_at(txn, coord),
-            Self::Reduce(reduce) => reduce.read_value_at(txn, coord),
-            Self::Slice(slice) => slice.read_value_at(txn, coord),
-            Self::Table(table) => table.read_value_at(txn, coord),
-            Self::Transpose(transpose) => transpose.read_value_at(txn, coord),
-            Self::Unary(unary) => unary.read_value_at(txn, coord),
-        }
+        dispatch!(self, this, this.read_value_at(txn, coord))
     }
 }
 
 impl<FD, FS, D, T> fmt::Display for SparseAccessor<FD, FS, D, T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match self {
-            Self::Broadcast(broadcast) => fmt::Display::fmt(broadcast, f),
-            Self::Cast(cast) => fmt::Display::fmt(cast, f),
-            Self::Combine(combinator) => fmt::Display::fmt(combinator, f),
-            Self::CombineConst(combinator) => fmt::Display::fmt(combinator, f),
-            Self::CombineLeft(combinator) => fmt::Display::fmt(combinator, f),
-            Self::Dense(dense) => fmt::Display::fmt(dense, f),
-            Self::Expand(expand) => fmt::Display::fmt(expand, f),
-            Self::Reduce(reduce) => fmt::Display::fmt(reduce, f),
-            Self::Slice(slice) => fmt::Display::fmt(slice, f),
-            Self::Table(table) => fmt::Display::fmt(table, f),
-            Self::Transpose(transpose) => fmt::Display::fmt(transpose, f),
-            Self::Unary(unary) => fmt::Display::fmt(unary, f),
-        }
+        dispatch!(self, this, fmt::Display::fmt(this, f))
     }
 }
 
@@ -925,7 +788,7 @@ where
     R: SparseAccess<FD, FS, D, T>,
 {
     fn dtype(&self) -> NumberType {
-        (self.combinator)(self.left.dtype().zero(), self.right.dtype().zero()).class()
+        (self.combinator)(self.left.dtype().one(), self.right.dtype().one()).class()
     }
 
     fn ndim(&self) -> usize {
@@ -1255,7 +1118,7 @@ where
     R: SparseAccess<FD, FS, D, T>,
 {
     fn dtype(&self) -> NumberType {
-        (self.combinator)(self.left.dtype().zero(), self.right.dtype().zero()).class()
+        (self.combinator)(self.left.dtype().one(), self.right.dtype().one()).class()
     }
 
     fn ndim(&self) -> usize {
@@ -1593,6 +1456,176 @@ impl<FD, FS, D, T, A> fmt::Display for SparseExpand<FD, FS, D, T, A> {
     }
 }
 
+#[derive(Clone)]
+pub struct SparseFlip<FD, FS, D, T, A> {
+    source: A,
+    rebase: transform::Flip,
+    phantom: Phantom<FD, FS, D, T>,
+}
+
+impl<FD, FS, D, T, A> SparseFlip<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    pub fn new(source: A, axis: usize) -> TCResult<Self> {
+        let rebase = transform::Flip::new(source.shape().clone(), axis)?;
+
+        Ok(Self {
+            source,
+            rebase,
+            phantom: Phantom::default(),
+        })
+    }
+}
+
+impl<FD, FS, D, T, A> TensorAccess for SparseFlip<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    fn dtype(&self) -> NumberType {
+        self.source.dtype()
+    }
+
+    fn ndim(&self) -> usize {
+        self.source.ndim()
+    }
+
+    fn shape(&'_ self) -> &'_ Shape {
+        self.source.shape()
+    }
+
+    fn size(&self) -> u64 {
+        self.source.size()
+    }
+}
+
+#[async_trait]
+impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseFlip<FD, FS, D, T, A>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Slice = SparseAccessor<FD, FS, D, T>;
+    type Transpose = SparseFlip<FD, FS, D, T, A::Transpose>;
+
+    fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
+        SparseAccessor::Flip(Box::new(SparseFlip {
+            source: self.source.accessor(),
+            rebase: self.rebase,
+            phantom: Phantom::default(),
+        }))
+    }
+
+    async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        let rebase = self.rebase;
+        let filled = self.source.filled(txn).await?;
+        let filled = filled.map_ok(move |(coord, value)| (rebase.flip_coord(coord), value));
+        Ok(Box::pin(filled))
+    }
+
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseExpand::filled_at {:?}", axes);
+
+        self.shape().validate_axes(&axes)?;
+
+        if axes.is_empty() {
+            return Ok(Box::pin(stream::empty()));
+        }
+
+        let shape = self.shape().clone();
+        let rebase = self.rebase;
+
+        let coords = self.source.filled_at(txn.clone(), axes).await?;
+        let coords = coords.map_ok(move |coords| rebase.flip_coords(coords));
+        let coords = sorted_coords::<FD, FS, D, T, _>(&txn, shape, coords).await?;
+        Ok(Box::pin(coords))
+    }
+
+    async fn filled_count(self, txn: T) -> TCResult<u64> {
+        self.source.filled_count(txn).await
+    }
+
+    fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseFlip {} slice {}", self.shape(), bounds);
+
+        self.shape().validate_bounds(&bounds)?;
+        bounds.normalize(self.shape());
+
+        if bounds == Bounds::all(self.shape()) {
+            return Ok(self.accessor());
+        }
+
+        if let Some(flip_axis) = self.rebase.invert_axis(&bounds) {
+            let source_bounds = self.rebase.flip_bounds(bounds);
+            let slice = self.source.slice(source_bounds)?;
+            SparseFlip::new(slice, flip_axis).map(|slice| slice.accessor())
+        } else {
+            let source_bounds = self.rebase.flip_bounds(bounds);
+            self.source
+                .slice(source_bounds)
+                .map(|slice| slice.accessor())
+        }
+    }
+
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        let flip_axis = if let Some(permutation) = &permutation {
+            if permutation.len() == self.ndim() {
+                permutation[self.rebase.axis()]
+            } else {
+                return Err(TCError::bad_request(
+                    "invalid permutation",
+                    permutation.iter().collect::<Tuple<&usize>>(),
+                ));
+            }
+        } else {
+            self.ndim() - self.rebase.axis()
+        };
+
+        let source = self.source.transpose(permutation)?;
+        SparseFlip::new(source, flip_axis)
+    }
+}
+
+impl<FD, FS, D, T, A> ReadValueAt<D> for SparseFlip<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Txn = T;
+
+    fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
+        let source_coord = self.rebase.flip_coord(coord.clone());
+        let read = self
+            .source
+            .read_value_at(txn, source_coord)
+            .map_ok(|(_, val)| (coord, val));
+
+        Box::pin(read)
+    }
+}
+
+impl<FD, FS, D, T, A> fmt::Display for SparseFlip<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a flipped sparse Tensor")
+    }
+}
+
 type Reductor<FD, FS, D, T> =
     fn(&SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>, T) -> TCBoxTryFuture<Number>;
 
@@ -1773,6 +1806,142 @@ impl<FD, FS, D, T> fmt::Display for SparseReduce<FD, FS, D, T> {
     }
 }
 
+#[derive(Clone)]
+pub struct SparseReshape<FD, FS, D, T, A> {
+    source: A,
+    rebase: transform::Reshape,
+    phantom: Phantom<FD, FS, D, T>,
+}
+
+impl<FD, FS, D, T, A> SparseReshape<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    pub fn new(source: A, shape: Shape) -> TCResult<Self> {
+        let rebase = transform::Reshape::new(source.shape().clone(), shape, "reshape")?;
+        Ok(Self {
+            source,
+            rebase,
+            phantom: Phantom::default(),
+        })
+    }
+}
+
+impl<FD, FS, D, T, A> TensorAccess for SparseReshape<FD, FS, D, T, A>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    fn dtype(&self) -> NumberType {
+        self.source.dtype()
+    }
+
+    fn ndim(&self) -> usize {
+        self.shape().len()
+    }
+
+    fn shape(&'_ self) -> &'_ Shape {
+        self.rebase.shape()
+    }
+
+    fn size(&self) -> u64 {
+        self.source.size()
+    }
+}
+
+#[async_trait]
+impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseReshape<FD, FS, D, T, A>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<TensorType>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Slice = SparseTable<FD, FS, D, T>;
+    type Transpose = SparseTranspose<FD, FS, D, T, Self>;
+
+    fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
+        let reshape = SparseReshape {
+            source: self.source.accessor(),
+            rebase: self.rebase,
+            phantom: self.phantom,
+        };
+
+        SparseAccessor::Reshape(Box::new(reshape))
+    }
+
+    async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!("SparseReshape::filled");
+
+        let rebase = self.rebase;
+        let source = self.source.filled(txn).await?;
+        Ok(Box::pin(source.map_ok(move |(coord, value)| {
+            (rebase.map_coord(coord), value)
+        })))
+    }
+
+    async fn filled_at<'a>(
+        self,
+        _txn: T,
+        _axes: Vec<usize>,
+    ) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        Err(TCError::unsupported(
+            "cannot slice a reshaped Tensor; make a copy first",
+        ))
+    }
+
+    async fn filled_count(self, txn: T) -> TCResult<u64> {
+        let axes = (0..self.ndim()).collect();
+        let filled = self.filled_at(txn, axes).await?;
+
+        filled
+            .try_fold(0u64, |count, _| future::ready(Ok(count + 1)))
+            .await
+    }
+
+    fn slice(self, _bounds: Bounds) -> TCResult<Self::Slice> {
+        Err(TCError::unsupported(
+            "cannot slice a reshaped Tensor; make a copy first",
+        ))
+    }
+
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        SparseTranspose::new(self, permutation)
+    }
+}
+
+impl<FD, FS, D, T, A> ReadValueAt<D> for SparseReshape<FD, FS, D, T, A>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Txn = T;
+
+    fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
+        self.source
+            .read_value_at(txn, self.rebase.invert_coord(coord))
+    }
+}
+
+impl<FD, FS, D, T, A> fmt::Display for SparseReshape<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a reshaped sparse Tensor")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseTranspose<FD, FS, D, T, A> {
     source: A,
