@@ -9,26 +9,31 @@ use destream::{de, en};
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, Stream, StreamExt, TryStreamExt};
 use log::debug;
-use safecast::{AsType, CastFrom};
+use safecast::{AsType, CastFrom, CastInto};
 
 use tc_btree::{BTreeType, Node};
 use tc_error::*;
-use tc_transact::fs::{CopyFrom, Dir, File, Hash, Persist, Restore};
+use tc_transact::fs::{CopyFrom, Dir, File, Persist, Restore};
 use tc_transact::{IntoView, Transact, Transaction, TxnId};
-use tc_value::{FloatType, Number, NumberClass, NumberInstance, NumberType};
-use tcgeneric::{Instance, TCBoxTryFuture, TCBoxTryStream};
+use tc_value::{
+    Float, FloatType, Number, NumberClass, NumberInstance, NumberType, Trigonometry, UIntType,
+};
+use tcgeneric::{Instance, TCBoxTryFuture};
 
 use super::dense::{BlockListSparse, DenseTensor, PER_BLOCK};
+use super::stream::ReadValueAt;
 use super::transform;
 use super::{
-    Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorBooleanConst,
-    TensorCompare, TensorCompareConst, TensorDualIO, TensorIO, TensorInstance, TensorMath,
-    TensorMathConst, TensorReduce, TensorTransform, TensorType, TensorUnary, ERR_COMPLEX_EXPONENT,
+    coord_bounds, tile, trig_dtype, AxisBounds, Bounds, Coord, Phantom, Schema, Shape, Tensor,
+    TensorAccess, TensorBoolean, TensorBooleanConst, TensorCompare, TensorCompareConst,
+    TensorDiagonal, TensorDualIO, TensorIO, TensorIndex, TensorInstance, TensorMath,
+    TensorMathConst, TensorPersist, TensorReduce, TensorTransform, TensorTrig, TensorType,
+    TensorUnary, ERR_COMPLEX_EXPONENT,
 };
 
-use crate::TensorPersist;
 use access::*;
 pub use access::{DenseToSparse, SparseAccess, SparseAccessor, SparseWrite};
+use combine::coord_to_offset;
 pub use table::SparseTable;
 
 mod access;
@@ -38,7 +43,7 @@ mod table;
 pub type SparseRow = (Coord, Number);
 pub type SparseStream<'a> = Pin<Box<dyn Stream<Item = TCResult<SparseRow>> + Send + Unpin + 'a>>;
 
-const ERR_NOT_SPARSE: &str = "The result of the requested operation would not be sparse;\
+const ERR_NOT_SPARSE: &str = "The result of the requested operation would not be sparse; \
 convert to a DenseTensor first.";
 
 /// A `Tensor` stored as a `Table` of [`Coord`]s and [`Number`] values
@@ -152,7 +157,7 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D::File: AsType<FD> + AsType<FS>,
-    D::FileClass: From<BTreeType>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
 {
     /// Create a new `SparseTensor` with the given schema
     pub async fn create(dir: &D, schema: Schema, txn_id: TxnId) -> TCResult<Self> {
@@ -160,6 +165,41 @@ where
             .map_ok(Self::from)
             .await
     }
+
+    /// Tile the given `tensor` into a new `SparseTensor`
+    pub async fn tile(
+        txn: T,
+        tensor: SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>,
+        multiples: Vec<u64>,
+    ) -> TCResult<Self> {
+        if multiples.len() != tensor.ndim() {
+            return Err(TCError::bad_request(
+                "wrong number of multiples to tile a Tensor with shape",
+                tensor.shape(),
+            ))?;
+        }
+
+        let txn_id = *txn.id();
+        let dir = txn.context().create_dir_unique(txn_id).await?;
+        let dtype = tensor.dtype();
+        let shape = tensor
+            .shape()
+            .iter()
+            .zip(&multiples)
+            .map(|(dim, m)| dim * m)
+            .collect();
+
+        let input = match tensor.accessor {
+            SparseAccessor::Table(table) => table.into(),
+            other => {
+                let dir = txn.context().create_dir_unique(*txn.id()).await?;
+                SparseTensor::copy_from(other.into(), dir, &txn).await?
+            }
+        };
+
+        let output = Self::create(&dir, Schema { shape, dtype }, txn_id).await?;
+        tile(txn, input, output, multiples).await
+    }
 }
 
 impl<FD, FS, D, T> TensorPersist for SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>> {
@@ -471,7 +511,69 @@ impl<FD, FS, D, T, A> TensorCompareConst for SparseTensor<FD, FS, D, T, A> {
 }
 
 #[async_trait]
-impl<FD, FS, D, T, L, R> TensorDualIO<D, SparseTensor<FD, FS, D, T, R>>
+impl<FD, FS, D, T, A> TensorDiagonal<D> for SparseTensor<FD, FS, D, T, A>
+where
+    D: Dir,
+    T: Transaction<D>,
+    FD: File<Array>,
+    FS: File<Node>,
+    A: SparseAccess<FD, FS, D, T>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
+    SparseTable<FD, FS, D, T>: ReadValueAt<D, Txn = T>,
+{
+    type Txn = T;
+    type Diagonal = SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>;
+
+    async fn diagonal(self, txn: Self::Txn) -> TCResult<Self::Diagonal> {
+        if self.ndim() != 2 {
+            return Err(TCError::not_implemented(format!(
+                "diagonal of a {}-dimensional sparse Tensor",
+                self.ndim()
+            )));
+        }
+
+        let size = self.shape()[0];
+        if size != self.shape()[1] {
+            return Err(TCError::bad_request(
+                "diagonal requires a square matrix but found",
+                self.shape(),
+            ));
+        }
+
+        let txn_id = *txn.id();
+        let dir = txn.context().create_dir_unique(txn_id).await?;
+
+        let shape = vec![size].into();
+        let dtype = self.dtype();
+        let schema = Schema { shape, dtype };
+        let table = SparseTable::create(&dir, schema, txn_id).await?;
+
+        let filled = self.accessor.filled(txn).await?;
+        filled
+            .try_filter_map(|(mut coord, value)| {
+                future::ready(Ok({
+                    debug_assert!(coord.len() == 2);
+                    debug_assert_ne!(value, value.class().zero());
+
+                    if coord.pop() == Some(coord[0]) {
+                        Some((coord, value))
+                    } else {
+                        None
+                    }
+                }))
+            })
+            .map_ok(|(coord, value)| table.write_value(txn_id, coord, value))
+            .try_buffer_unordered(num_cpus::get())
+            .try_fold((), |(), ()| future::ready(Ok(())))
+            .await?;
+
+        Ok(table.into())
+    }
+}
+
+#[async_trait]
+impl<FD, FS, D, T, L> TensorDualIO<D, SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>>
     for SparseTensor<FD, FS, D, T, L>
 where
     D: Dir,
@@ -481,7 +583,6 @@ where
     D::File: AsType<FD> + AsType<FS>,
     D::FileClass: From<TensorType>,
     L: SparseWrite<FD, FS, D, T>,
-    R: SparseAccess<FD, FS, D, T>,
 {
     type Txn = T;
 
@@ -489,7 +590,7 @@ where
         self,
         txn: T,
         bounds: Bounds,
-        other: SparseTensor<FD, FS, D, T, R>,
+        other: SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>,
     ) -> TCResult<()> {
         let slice_shape = bounds.to_shape(self.shape())?;
         if &slice_shape != other.shape() {
@@ -524,7 +625,7 @@ where
     FD: File<Array>,
     FS: File<Node>,
     D::File: AsType<FD> + AsType<FS>,
-    D::FileClass: From<TensorType>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
     A: SparseWrite<FD, FS, D, T>,
 {
     type Txn = T;
@@ -543,12 +644,151 @@ where
         };
 
         match other {
-            Tensor::Dense(other) => self.write(txn, bounds, other.into_sparse()).await,
-            Tensor::Sparse(other) => self.write(txn, bounds, other).await,
+            Tensor::Dense(other) => {
+                let dir = txn.context().create_dir_unique(*txn.id()).await?;
+                let other = SparseTensor::copy_from(other.into_sparse(), dir, &txn).await?;
+
+                self.write(txn, bounds, other.into_sparse()).await
+            }
+            Tensor::Sparse(other) => match other.accessor {
+                SparseAccessor::Table(table) => {
+                    self.write(txn, bounds, SparseTensor::from(table)).await
+                }
+                other => {
+                    let dir = txn.context().create_dir_unique(*txn.id()).await?;
+                    let other = SparseTensor::copy_from(other.into(), dir, &txn).await?;
+                    self.write(txn, bounds, other).await
+                }
+            },
         }
     }
 }
 
+#[async_trait]
+impl<FD, FS, D, T, A> TensorIndex<D> for SparseTensor<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseWrite<FD, FS, D, T>,
+    D::File: AsType<FD> + AsType<FS>,
+    D::FileClass: From<BTreeType> + From<TensorType>,
+{
+    type Txn = T;
+    type Index = SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>;
+
+    async fn argmax(self, txn: Self::Txn, axis: usize) -> TCResult<Self::Index> {
+        if axis >= self.ndim() {
+            return Err(TCError::unsupported(format!(
+                "invalid argmax axis for tensor with {} dimensions: {}",
+                self.ndim(),
+                axis
+            )));
+        }
+
+        let shape = {
+            let mut shape = self.shape().clone();
+            shape.remove(axis);
+            shape
+        };
+
+        let schema = Schema {
+            shape,
+            dtype: NumberType::UInt(UIntType::U64),
+        };
+
+        let txn_id = *txn.id();
+        let dir = txn.context().create_dir_unique(txn_id).await?;
+        let table = SparseTable::create(&dir, schema, txn_id).await?;
+
+        let dim = self.shape()[axis];
+        let zero = self.dtype().zero();
+        let axes = (0..self.ndim())
+            .into_iter()
+            .filter(|x| x != &axis)
+            .collect();
+
+        let mut filled = self.accessor.clone().filled_at(txn.clone(), axes).await?;
+        while let Some(coords) = filled.try_next().await? {
+            for coord in coords.to_vec() {
+                let mut bounds: Bounds = coord.iter().cloned().map(AxisBounds::At).collect();
+                bounds.insert(axis, AxisBounds::all(dim));
+
+                let slice = self.accessor.clone().slice(bounds)?;
+                debug_assert_eq!(slice.ndim(), 1);
+
+                let filled = slice.filled(txn.clone()).await?;
+                let filled = filled.map_ok(|(offset, value)| (offset[0], value));
+                let imax = imax(filled, zero, dim).await?;
+                table.write_value(txn_id, coord, imax.0.into()).await?;
+            }
+        }
+
+        Ok(table.into())
+    }
+
+    async fn argmax_all(self, txn: Self::Txn) -> TCResult<u64> {
+        let zero = self.dtype().zero();
+        let size = self.size();
+        let coord_bounds = coord_bounds(self.shape());
+        let filled = self.accessor.filled(txn).await?;
+        let filled =
+            filled.map_ok(move |(coord, value)| (coord_to_offset(&coord, &coord_bounds), value));
+        let imax = imax(filled, zero, size).await?;
+        Ok(imax.0)
+    }
+}
+
+async fn imax<F>(mut filled: F, zero: Number, size: u64) -> TCResult<(u64, Number)>
+where
+    F: Stream<Item = TCResult<(u64, Number)>> + Unpin,
+{
+    let mut first_empty = Some(0);
+    let mut last = 0u64;
+    let mut imax = None;
+    while let Some((offset, value)) = filled.try_next().await? {
+        if offset == 0 {
+            first_empty = None;
+        } else if first_empty.is_none() {
+            if offset > (last + 1) {
+                first_empty = Some(last + 1)
+            }
+        }
+
+        if let Some((ref mut i, ref mut max)) = &mut imax {
+            if value > *max {
+                *i = offset;
+                *max = value;
+            }
+        } else {
+            imax = Some((offset, value));
+        }
+
+        last = offset;
+    }
+
+    if first_empty.is_none() && last < (size - 1) {
+        if last == 0 && imax.is_none() {
+            first_empty = Some(0);
+        } else {
+            first_empty = Some(last + 1);
+        }
+    }
+
+    if let Some((i, max)) = imax {
+        if max > zero {
+            Ok((i, max))
+        } else if let Some(first_empty) = first_empty {
+            Ok((first_empty, zero))
+        } else {
+            Ok((i, max))
+        }
+    } else {
+        Ok((0, zero))
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T, A> TensorIO<D> for SparseTensor<FD, FS, D, T, A>
 where
@@ -606,7 +846,28 @@ where
 
     fn div(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
         debug!("SparseTensor::div");
-        self.left_combine(other, Number::div)
+        fn div(l: Number, r: Number) -> Number {
+            // to prevent a divide-by-zero error, treat the right-hand side as if it doesn't exist
+            if r == r.class().zero() {
+                Ord::max(l.class(), r.class()).zero()
+            } else {
+                l / r
+            }
+        }
+
+        self.left_combine(other, div)
+    }
+
+    fn log(self, base: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
+        if base.dtype().is_complex() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
+
+        fn log(n: Number, base: Number) -> Number {
+            n.log(Float::cast_from(base))
+        }
+
+        self.left_combine(base, log)
     }
 
     fn mul(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
@@ -615,7 +876,7 @@ where
     }
 
     fn pow(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
-        if !other.dtype().is_real() {
+        if other.dtype().is_complex() {
             return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
         }
 
@@ -656,6 +917,13 @@ where
         }
     }
 
+    fn log(self, base: Tensor<FD, FS, D, T>) -> TCResult<Self::LeftCombine> {
+        match base {
+            Tensor::Sparse(sparse) => self.log(sparse).map(Tensor::from),
+            Tensor::Dense(dense) => self.log(dense.into_sparse()).map(Tensor::from),
+        }
+    }
+
     fn mul(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
             Tensor::Sparse(sparse) => self.mul(sparse).map(Tensor::from),
@@ -689,6 +957,23 @@ impl<FD, FS, D, T, A> TensorMathConst for SparseTensor<FD, FS, D, T, A> {
         Ok(SparseConstCombinator::new(self.accessor, other, Number::div).into())
     }
 
+    fn log_const(self, base: Number) -> TCResult<Self::Combine> {
+        if base.class().is_complex() {
+            return Err(TCError::unsupported(ERR_COMPLEX_EXPONENT));
+        }
+
+        fn log(n: Number, base: Number) -> Number {
+            if let Number::Float(base) = base {
+                n.log(base)
+            } else {
+                unreachable!("log with non-floating point base")
+            }
+        }
+
+        let base = Number::Float(base.cast_into());
+        Ok(SparseConstCombinator::new(self.accessor, base, log).into())
+    }
+
     fn mul_const(self, other: Number) -> TCResult<Self::Combine> {
         Ok(SparseConstCombinator::new(self.accessor, other, Number::mul).into())
     }
@@ -781,6 +1066,8 @@ where
     type Broadcast = SparseTensor<FD, FS, D, T, SparseBroadcast<FD, FS, D, T, A>>;
     type Cast = SparseTensor<FD, FS, D, T, SparseCast<FD, FS, D, T, A>>;
     type Expand = SparseTensor<FD, FS, D, T, SparseExpand<FD, FS, D, T, A>>;
+    type Flip = SparseTensor<FD, FS, D, T, SparseFlip<FD, FS, D, T, A>>;
+    type Reshape = SparseTensor<FD, FS, D, T, SparseReshape<FD, FS, D, T, A>>;
     type Slice = SparseTensor<FD, FS, D, T, A::Slice>;
     type Transpose = SparseTensor<FD, FS, D, T, A::Transpose>;
 
@@ -799,6 +1086,16 @@ where
         Ok(accessor.into())
     }
 
+    fn flip(self, axis: usize) -> TCResult<Self::Flip> {
+        let accessor = SparseFlip::new(self.accessor, axis)?;
+        Ok(accessor.into())
+    }
+
+    fn reshape(self, shape: Shape) -> TCResult<Self::Reshape> {
+        let accessor = SparseReshape::new(self.accessor, shape)?;
+        Ok(accessor.into())
+    }
+
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         let accessor = self.accessor.slice(bounds)?;
         Ok(accessor.into())
@@ -810,6 +1107,44 @@ where
     }
 }
 
+macro_rules! trig {
+    ($fun:ident) => {
+        fn $fun(&self) -> TCResult<Self::Unary> {
+            let dtype = trig_dtype(self.dtype());
+            let source = self.accessor.clone().accessor();
+            let accessor = SparseUnary::new(source, Number::$fun, dtype);
+            Ok(SparseTensor::from(accessor))
+        }
+    };
+}
+
+#[async_trait]
+impl<FD, FS, D, T, A> TensorTrig for SparseTensor<FD, FS, D, T, A>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    A: SparseAccess<FD, FS, D, T>,
+{
+    type Unary = SparseTensor<FD, FS, D, T, SparseUnary<FD, FS, D, T>>;
+
+    trig! {asin}
+    trig! {sin}
+    trig! {sinh}
+    trig! {asinh}
+
+    trig! {acos}
+    trig! {cos}
+    trig! {cosh}
+    trig! {acosh}
+
+    trig! {atan}
+    trig! {tan}
+    trig! {tanh}
+    trig! {atanh}
+}
+
 #[async_trait]
 impl<FD, FS, D, T, A> TensorUnary<D> for SparseTensor<FD, FS, D, T, A>
 where
@@ -826,7 +1161,7 @@ where
         let source = self.accessor.clone().accessor();
         let transform = <Number as NumberInstance>::abs;
 
-        let accessor = SparseUnary::new(source, transform, self.dtype());
+        let accessor = SparseUnary::new(source, transform, self.dtype().one().abs().class());
         Ok(SparseTensor::from(accessor))
     }
 
@@ -842,6 +1177,20 @@ where
         Ok(SparseTensor::from(accessor))
     }
 
+    fn ln(&self) -> TCResult<Self::Unary> {
+        let dtype = self.dtype().one().ln().class();
+        let source = self.accessor.clone().accessor();
+        let accessor = SparseUnary::new(source, Number::ln, dtype);
+        Ok(SparseTensor::from(accessor))
+    }
+
+    fn round(&self) -> TCResult<Self::Unary> {
+        let dtype = self.dtype().one().ln().class();
+        let source = self.accessor.clone().accessor();
+        let accessor = SparseUnary::new(source, Number::round, dtype);
+        Ok(SparseTensor::from(accessor))
+    }
+
     async fn all(self, txn: Self::Txn) -> TCResult<bool> {
         let affected = stream::iter(Bounds::all(self.shape()).affected());
         let filled = self.accessor.filled(txn).await?;
@@ -893,23 +1242,6 @@ where
     }
 }
 
-#[async_trait]
-impl<'en, FD, FS, D, T, A> Hash<'en, D> for SparseTensor<FD, FS, D, T, A>
-where
-    FD: File<Array>,
-    FS: File<Node>,
-    D: Dir,
-    T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
-{
-    type Item = SparseRow;
-    type Txn = T;
-
-    async fn hashable(&'en self, txn: &'en Self::Txn) -> TCResult<TCBoxTryStream<'en, SparseRow>> {
-        self.accessor.clone().filled(txn.clone()).await
-    }
-}
-
 #[async_trait]
 impl<FD, FS, D, T> Persist<D> for SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
 where
@@ -1057,8 +1389,9 @@ where
     }
 
     async fn visit_seq<A: de::SeqAccess>(self, mut seq: A) -> Result<Self::Value, A::Error> {
-        let schema = seq.next_element(()).await?;
+        let schema = seq.next_element::<Schema>(()).await?;
         let schema = schema.ok_or_else(|| de::Error::invalid_length(0, "tensor schema"))?;
+        schema.validate("load Sparse").map_err(de::Error::custom)?;
 
         let txn_id = *self.txn.id();
         let table = SparseTable::create(self.txn.context(), schema, txn_id)
