@@ -1,5 +1,6 @@
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::iter;
+use std::ops;
 
 use afarray::Coords;
 use log::debug;
@@ -22,6 +23,9 @@ pub struct Broadcast {
 
 impl Broadcast {
     pub fn new(source_shape: Shape, shape: Shape) -> TCResult<Broadcast> {
+        source_shape.validate("broadcast")?;
+        shape.validate("broadcast")?;
+
         if source_shape.is_empty() {
             return Err(TCError::unsupported("cannot broadcast an empty Tensor"));
         } else if shape.is_empty() {
@@ -256,6 +260,70 @@ impl Expand {
     }
 }
 
+#[derive(Clone)]
+pub struct Flip {
+    shape: Shape,
+    axis: usize,
+}
+
+impl Flip {
+    pub fn new(shape: Shape, axis: usize) -> TCResult<Self> {
+        if axis > shape.len() {
+            Err(TCError::unsupported(format!(
+                "invalid axis {} for shape {}",
+                axis, shape
+            )))
+        } else {
+            Ok(Self { shape, axis })
+        }
+    }
+
+    pub fn axis(&self) -> usize {
+        self.axis
+    }
+
+    pub fn flip_bounds(&self, mut bounds: Bounds) -> Bounds {
+        if bounds.len() < self.axis {
+            return bounds;
+        }
+
+        let dim = self.shape[self.axis];
+        bounds[self.axis] = match &bounds[self.axis] {
+            AxisBounds::At(i) => AxisBounds::At(dim - i),
+            AxisBounds::In(ops::Range { start, end }) => AxisBounds::In((dim - end)..(dim - start)),
+            AxisBounds::Of(indices) => {
+                AxisBounds::Of(indices.into_iter().map(|i| dim - i).collect())
+            }
+        };
+
+        bounds
+    }
+
+    pub fn flip_coord(&self, mut coord: Coord) -> Coord {
+        assert_eq!(coord.len(), self.shape.len());
+        coord[self.axis] = self.shape[self.axis] - coord[self.axis];
+        coord
+    }
+
+    pub fn flip_coords(&self, coords: Coords) -> Coords {
+        coords.flip(&self.shape, self.axis)
+    }
+
+    pub fn invert_axis(&self, bounds: &Bounds) -> Option<usize> {
+        if bounds.len() <= self.axis {
+            None
+        } else if bounds[self.axis].is_index() {
+            None
+        } else {
+            let elided = bounds[..self.axis]
+                .iter()
+                .filter(|bound| bound.is_index())
+                .count();
+            Some(self.axis - elided)
+        }
+    }
+}
+
 #[derive(Clone)]
 pub struct Reduce {
     source_shape: Shape,
@@ -265,11 +333,19 @@ pub struct Reduce {
 
 impl Reduce {
     pub fn new(source_shape: Shape, axis: usize) -> TCResult<Reduce> {
+        if source_shape.size() == 0 {
+            return Err(TCError::unsupported("cannot reduce a zero-size tensor"));
+        }
         if axis >= source_shape.len() {
             return Err(TCError::unsupported(format!(
                 "cannot reduce axis {} of tensor with shape {}",
                 axis, source_shape
             )));
+        } else if source_shape[axis] == 0 {
+            return Err(TCError::unsupported(format!(
+                "cannot reduce axis {} with dimension 0",
+                axis
+            )));
         }
 
         let mut shape = source_shape.clone();
@@ -316,11 +392,93 @@ impl Reduce {
         bounds.into()
     }
 
+    pub fn invert_coords(&self, coords: Coords) -> Coords {
+        coords.expand(&self.source_shape, self.axis)
+    }
+
     pub fn reduce_axis(&self) -> usize {
         self.axis
     }
 }
 
+#[derive(Clone)]
+pub struct Reshape {
+    source_shape: Shape,
+    source_bounds: Vec<u64>,
+
+    shape: Shape,
+    bounds: Vec<u64>,
+}
+
+impl Reshape {
+    pub fn new(source_shape: Shape, shape: Shape, debug_info: &'static str) -> TCResult<Self> {
+        source_shape.validate(debug_info)?;
+        shape.validate(debug_info)?;
+
+        if source_shape.size() != shape.size() {
+            return Err(TCError::unsupported(format!(
+                "cannot reshape tensor with shape {} into {}",
+                source_shape, shape
+            )));
+        }
+
+        let source_bounds = coord_bounds(&source_shape);
+        let bounds = coord_bounds(&shape);
+
+        Ok(Self {
+            source_shape,
+            source_bounds,
+            shape,
+            bounds,
+        })
+    }
+
+    pub fn shape(&self) -> &Shape {
+        &self.shape
+    }
+
+    pub fn invert_coord(&self, coord: Coord) -> Coord {
+        assert_eq!(coord.len(), self.shape.len());
+
+        let offset: u64 = coord
+            .into_iter()
+            .zip(&self.bounds)
+            .map(|(x, bound)| x * bound)
+            .sum();
+
+        self.source_bounds
+            .iter()
+            .map(|bound| offset / bound)
+            .zip(self.source_shape.iter())
+            .map(|(axis_offset, dim)| axis_offset % dim)
+            .collect()
+    }
+
+    pub fn invert_coords(&self, coords: Coords) -> Coords {
+        assert_eq!(coords.ndim(), self.shape.len());
+
+        let offsets = coords.to_offsets(&self.shape);
+        Coords::from_offsets(offsets, &self.source_shape)
+    }
+
+    pub fn map_coord(&self, coord: Coord) -> Coord {
+        assert_eq!(coord.len(), self.source_shape.len());
+
+        let offset: u64 = coord
+            .into_iter()
+            .zip(&self.source_bounds)
+            .map(|(x, bound)| x * bound)
+            .sum();
+
+        self.bounds
+            .iter()
+            .map(|bound| offset / bound)
+            .zip(self.shape.iter())
+            .map(|(axis_offset, dim)| axis_offset % dim)
+            .collect()
+    }
+}
+
 #[derive(Clone)]
 pub struct Slice {
     source_shape: Shape,
@@ -497,7 +655,12 @@ pub struct Transpose {
 impl Transpose {
     pub fn new(source_shape: Shape, permutation: Option<Vec<usize>>) -> TCResult<Transpose> {
         let ndim = source_shape.len();
-        let permutation = permutation.unwrap_or_else(|| (0..ndim).rev().collect());
+
+        let permutation = if let Some(permutation) = permutation {
+            permutation
+        } else {
+            (0..ndim).rev().collect()
+        };
 
         if permutation.len() != ndim {
             return Err(TCError::unsupported(format!(
@@ -505,19 +668,23 @@ impl Transpose {
                 source_shape,
                 Tuple::from(permutation)
             )));
-        } else if let Some(max_axis) = permutation.iter().max() {
-            if max_axis >= &ndim {
-                return Err(TCError::bad_request(
-                    "cannot transpose nonexistent axis",
-                    max_axis,
-                ));
-            }
+        } else if permutation.iter().max().expect("transpose last axis") > &ndim {
+            return Err(TCError::bad_request(
+                "cannot transpose nonexistent axis",
+                permutation.iter().max().unwrap(),
+            ));
+        } else if permutation.iter().cloned().collect::<HashSet<_>>().len() != permutation.len() {
+            return Err(TCError::bad_request(
+                "cannot transpose the same axis twice",
+                Tuple::from(permutation),
+            ));
         }
 
         let mut shape: Coord = Vec::with_capacity(ndim);
         for axis in &permutation {
             shape.push(source_shape[*axis]);
         }
+
         let shape: Shape = shape.into();
 
         let mut inverse_permutation = vec![0; ndim];
@@ -609,6 +776,13 @@ impl Transpose {
     }
 }
 
+#[inline]
+fn coord_bounds(shape: &[u64]) -> Vec<u64> {
+    (0..shape.len())
+        .map(|axis| shape[axis + 1..].iter().product())
+        .collect()
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -623,6 +797,16 @@ mod tests {
         )
     }
 
+    #[test]
+    fn test_reshape() {
+        let source = Shape::from(vec![2, 3, 4, 1]);
+        let dest = Shape::from(vec![3, 8]);
+        let rebase = Reshape::new(source, dest).unwrap();
+
+        assert_eq!(rebase.map_coord(vec![0, 1, 2, 0]), vec![0, 6]);
+        assert_eq!(rebase.invert_coord(vec![0, 6]), vec![0, 1, 2, 0]);
+    }
+
     #[test]
     fn test_slice_invert_bounds() {
         let rebase = Slice::new(vec![2, 3, 4, 5].into(), Bounds::from(vec![0])).unwrap();
